package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.module.JFedWebApiClientModule;
import be.iminds.ilabt.jfed.call_log_output.LogOutput;
import be.iminds.ilabt.jfed.espec.bundle.BundleFetcher;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.filefetcher.FileFetcher;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ExperimentSpecificationParser;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.UploadProgressTracker;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.ExperimentControllerFactory;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupConfig;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedExperimenterGuiConfigProvider;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfigImpl;
import be.iminds.ilabt.jfed.experimenter_gui.config.UserInfoProvider;
import be.iminds.ilabt.jfed.util.lib.ConnectivityDetector;
import be.iminds.ilabt.jfed.experimenter_gui.config.util.GuiConfigRSpecGenerator;
import be.iminds.ilabt.jfed.experimenter_gui.config.util.GuiConfigStitchingTestRSpecGenerator;
import be.iminds.ilabt.jfed.experimenter_gui.config.util.TestbedNodesMapsFetcher;
import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClient;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.git.SingleSshGitAuthPreferences;
import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.highlevel.LowLevelModule;
import be.iminds.ilabt.jfed.highlevel.SfaOnlyExperimentModule;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.ApiInfo;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.finder.AuthorityFinder;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedCorePreferences;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGenerator;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.rspec.generator.StitchingTestRSpecGenerator;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.testing.base.*;
import be.iminds.ilabt.jfed.testing.shared.CommonAMTest;
import be.iminds.ilabt.jfed.testing.shared.NodeLoginTestStep;
import be.iminds.ilabt.jfed.testing.shared.Proxy;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.highlevel.util.JavaFXLogger;
import be.iminds.ilabt.jfed.util.library.KeyUtil;
import be.iminds.ilabt.jfed.lowlevel.authority.legacy.TargetAuthority;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.inject.*;
import com.google.inject.Module;
import com.google.inject.name.Names;
import com.google.inject.util.Modules;
import io.dropwizard.jackson.Jackson;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import java.io.IOException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.security.NoSuchAlgorithmException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.stream.Collectors;

/**
 * GuiLogicTest
 */
public class ESpecTest extends ApiTest<ESpecTestConfig> {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ESpecTest.class);

    @Nonnull private final Server userAuthorityServer;
    @Nonnull private final TestbedInfoSource testbedInfoSource;
    @Nonnull private final AuthorityFinder authorityFinder;

    @Nonnull private final Injector originalInjector;
    @Nonnull private final JFedCorePreferences configJFedPreferences;
    @Nonnull private final Injector configInjector;

    private final ESpecLogToTest eSpecLogToTest;

    @Inject
    public ESpecTest(@Nonnull Injector originalInjector,
                     @Nonnull Logger logger,
                     @Nullable TargetAuthority testedAuthority,
                     @Nonnull GeniUserProvider geniUserProvider,
                     @Nonnull ESpecTestConfig testConfig) {
        super(logger, testedAuthority, geniUserProvider,
                testConfig);

        this.originalInjector = originalInjector;
        this.configJFedPreferences = setupPreferences();
        this.configInjector = createInjector(configJFedPreferences);

        this.testbedInfoSource = configInjector.getInstance(TestbedInfoSource.class);
        this.authorityFinder = configInjector.getInstance(AuthorityFinder.class);

        if (user == null)
            throw new IllegalStateException("User is null");

        Server uas = user.getUserAuthorityServer();
        if (uas == null)
            throw new IllegalStateException("User has no Authority Server");

        userAuthorityServer = uas;

        this.eSpecLogToTest = new ESpecLogToTest();
    }

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();
    private static final ApiTestMetaData metadata = new ApiTestMetaData<ESpecTestConfig>() {
        @Nonnull
        @Override
        public String getTestDescription() {
            return "A test using the jFed GUI Editor logic. This will reserve resources, test login, and delete them again.";
        }

        @Nonnull
        @Override
        public Class<ESpecTestConfig> getConfigClass() {
            return ESpecTestConfig.class;
        }

        @Nonnull
        @Override
        public ESpecTestConfig parseApiTestConfig(@Nonnull String apiTestConfigString) {
            try {
                return MAPPER.readValue(apiTestConfigString, ESpecTestConfig.class);
            } catch (IOException e) {
                throw new RuntimeException("Invalid Test config", e);
            }
        }

        @Nonnull
        @Override
        public ApiTestConfigEditor<ESpecTestConfig> getApiTestConfigEditor() {
            return new JsonTestConfigEditor(ESpecTestConfig.class);
        }

    };
    public static ApiTestMetaData getMetaData() {
        return metadata;
    }

    private JFedGuiConfig cachedGuiConfig = null;
    private JFedGuiConfig getGuiConfig() {
        if (cachedGuiConfig != null)
            return cachedGuiConfig;

        ConnectivityDetector connectivityDetector = configInjector.getInstance(ConnectivityDetector.class);
        UserInfoProvider userInfoProvider = configInjector.getInstance(UserInfoProvider.class);
        URL webApiUrl =
                configInjector.getInstance(Key.get(URL.class, Names.named(JFedWebApiClientModule.NAMED__NO_CLIENT_AUTH_WEBAPI_URL)));
        FedmonWebApiClient fedmonWebApiClient = configInjector.getInstance(FedmonWebApiClient.class);

        if (geniUserProvider == null) {
            throw new NullPointerException("geniUserProvider may not be null");
        }

        JFedExperimenterGuiConfigProvider jFedExperimenterGuiConfigProvider =
                new JFedExperimenterGuiConfigProvider(
                        geniUserProvider,
                        configJFedPreferences,
                        connectivityDetector,
                        userInfoProvider,
                        webApiUrl,
                        logger,
                        false);
        TestbedNodesMapsFetcher testbedNodesMapsFetcher =
                new TestbedNodesMapsFetcher(connectivityDetector);
        cachedGuiConfig = new JFedGuiConfigImpl(
                testbedInfoSource,
                authorityFinder,
                jFedExperimenterGuiConfigProvider,
                testbedNodesMapsFetcher,
                fedmonWebApiClient);
        return cachedGuiConfig;
    }

    private RSpecGeneratorFactory getRSpecGeneratorFactory() {
        return new RSpecGeneratorFactory() {
            @Nonnull
            @Override
            public RSpecGenerator createRSpecGenerator() {
                return new GuiConfigRSpecGenerator(getGuiConfig(), authorityFinder, testbedInfoSource);
            }

            @Nonnull
            @Override
            public StitchingTestRSpecGenerator createStitchingTestRSpecGenerator() {
                return new GuiConfigStitchingTestRSpecGenerator(getGuiConfig(), authorityFinder, testbedInfoSource);
            }
        };
    }

    private JFedCorePreferences setupPreferences() {
        //setup proxy correctly by changing original JFedPreferences
        JFedPreferences basePrefs = originalInjector.getInstance(JFedPreferences.class);
        JFedCorePreferences newJFedPreferences = basePrefs instanceof JFedCorePreferences ?
                ((JFedCorePreferences) basePrefs).makeNoSaveCopy() :
                null;
        if (newJFedPreferences != null) {
            if (getTestConfig().getAmConnectionProxy().getMode().equals(Proxy.ProxyMode.MANUAL))
                throw new RuntimeException("Option amConnectionProxy.mode == "+Proxy.ProxyMode.MANUAL+" is not supported");

            newJFedPreferences.setString(
                    CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED,
                    getTestConfig().getAmConnectionProxy().getMode().equals(Proxy.ProxyMode.USER_AUTHORITY) ?
                            "ALWAYS" : "NEVER"
            );
            LOG.debug("Set usage of \"call\" proxy to: "+
                    newJFedPreferences.getString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED));
//            newJFedPreferences.setString(
//                    CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH,
//                    getTestConfig().getNodeLoginTest().getProxy().getMode().equals(Proxy.ProxyMode.USER_AUTHORITY) ?
//                            "ALWAYS" : "NEVER"
//            );
//            note("Set usage of SSH proxy to: "+
//                    newJFedPreferences.getString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH));
        }
        return newJFedPreferences;
    }


    private Injector createInjector(JFedCorePreferences configPreferences) {
        //just create an injector specifically for us: Do not share this stuff with any other tests that might be around.
        List<Module> minimalModules = new ArrayList<>();
        minimalModules.add(new LowLevelModule());
        minimalModules.add(new HighLevelModule());
        minimalModules.add(new SfaOnlyExperimentModule());
        minimalModules.add(binder -> binder.bind(GeniUserProvider.class).toInstance(geniUserProvider));
        minimalModules.add(binder -> binder.bind(JFedPreferences.class).toInstance(configPreferences));
        minimalModules.add(binder -> binder.bind(JFedCorePreferences.class).toInstance(configPreferences));

        //make sure correct logger is used
        AbstractModule loggerOverride = new AbstractModule() {
            @Override
            protected void configure() {
                bind(Logger.class).toInstance(getLogger());
                if (getLogger() instanceof JavaFXLogger) {
                    bind(JavaFXLogger.class).toInstance((JavaFXLogger) getLogger());
                } else {
                    //if it's not, that's a problem.
                    // You might run into problems with non-singleton logger later on, which results in calls not being logged.
                    throw new RuntimeException("Expected getLogger() to be a JavaFXLogger.");
                }
            }
        };

        Injector resInjector = Guice.createInjector(Modules.override(minimalModules).with(loggerOverride));

        assertNotNull(resInjector);
        LOG.debug("Created injector");

        Logger injectorLogger = resInjector.getInstance(Logger.class);
        //noinspection ObjectEquality
        if (getLogger() != injectorLogger) {
            errorFatal("Invalid logger in injector");
        } else {
            LOG.debug("Logger OK");
        }

        JFedPreferences injectorPreferences = resInjector.getInstance(JFedPreferences.class);
        //noinspection ObjectEquality
        if (configPreferences != injectorPreferences) {
            errorFatal("Invalid JFedPreferences in injector");
        } else {
            LOG.debug("JFedPreferences OK");
        }

        return resInjector;
    }

    @Override
    public void setUp() throws Exception {
        assertNotNull(geniUserProvider, "No user specified for test");
        assertNotNull(user, "No user specified for test");
        assertNotNull(getTestedAuthority(), "No target server specified for test");
        assertNotNull(getTestConfig(), "no test config");
        assertTrue(getTestConfig().isValid(), "Test config is not valid");

        assertTrue(getTestConfig().getProvision().isEnabled(),
                "A disabled provision step is currently not supported.");

    }




    private GeniUrn sliceUrn;

    private int maxNodeNameLength = 5; //TODO: automatically get from rspec
    private List<Server> serversInRspec;
    public String sliceName;

    private String project;
    private Experiment experiment;
    private boolean notEnoughResourcesDetected = false;
    private ExperimentController experimentController;

    private boolean waitForExperimentIsOver = false;
    private boolean deleted = false;

    String manifestRspecString;
    boolean validManifest;


    private ESpecBundle eSpecBundle;
    private ExperimentSpecification eSpec;



    @Test(groups = {"init"})
    public void getEspec() throws GeniUrn.GeniUrnParseException, IOException, ExperimentSpecificationParser.ExperimentSpecificationParseException, ESpecBundle.ESpecBundleInitException, InterruptedException {
        ExperimentSetupConfig.ESpec eConfig = getTestConfig().getESpec();

        assertNotNull(eConfig.getProvidedContentSource());

        switch (eConfig.getSource()) {
            case PROVIDE_ARCHIVE_FILE: {
                eSpecBundle = BundleFetcher.fetchArchiveFile(eConfig.getProvidedContentSource());
                break;
            }
            case PROVIDE_DIR: {
                eSpecBundle = BundleFetcher.fetchDir(eConfig.getProvidedContentSource());
                break;
            }
            case PROVIDE_ARCHIVE_URL: {
                eSpecBundle = BundleFetcher.fetchArchiveUrl(eConfig.getProvidedContentSource());
                break;
            }
            case PROVIDE_GIT_REPO_DIR: {
                eSpecBundle = BundleFetcher.fetchGitRepoDir(eConfig.getProvidedContentSource());
                break;
            }
            case DIRECT: {
                eSpecBundle = BundleFetcher.fromDirect(eConfig.getProvidedContentSource());
                break;
            }
            default: throw new IllegalStateException("Unsupported espec.source: "+eConfig.getSource().name());
        }
        assertNotNull(eSpecBundle);

        currentTestResult.addExtraResult(
                new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                        .setKey(ESpecBundle.ESPEC_DESCRIPTION_FILENAME)
                        .setValue(eSpecBundle.getExperimentSpecificationYml())
                        .setLarge(true)
                        .setHttpMediaTypeToXml()
        );

        ExperimentSpecificationParser parser = new ExperimentSpecificationParser();
        eSpec = parser.parse(eSpecBundle.getExperimentSpecificationYml());

        @SuppressWarnings("ConstantConditions")
        String userPem = user == null || user.getPrivateKey() == null ? null : new String(KeyUtil.privateKeyToAnyPem(user.getPrivateKey()));

        String requestRspec =
                new String(new FileFetcher(eSpec.getRspecs().get(0).getSource(),
                        eSpecBundle, getRSpecGeneratorFactory(), new SingleSshGitAuthPreferences(userPem)).fetchBytes(),
                        StandardCharsets.UTF_8);
        RequestRspecSource requestRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
        serversInRspec = requestRspecSource.getAllComponentManagerUrns().stream()
                .map(serverUrn ->
                        authorityFinder.findByUrn(
                                serverUrn,
                                AuthorityFinder.Purpose.REQUEST_RSPEC)
                )
                .collect(Collectors.toList());

        maxNodeNameLength = requestRspecSource.getBasicNodeInfo().stream()
                .map(BasicStringRspec.BasicNodeInfo::getClientId)
                .filter(Objects::nonNull)
                .mapToInt(String::length)
                .max()
                .orElse(5); //uses 5 as fallback when no nodes with client_id found

        currentTestResult.addExtraResult(
                new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                        .setKey("rspec")
                        .setValue(requestRspec)
                        .setLarge(true)
                        .setHttpMediaTypeToXml()
        );
    }

    @Nullable
    private ExperimentSetupConfig.ESpecLogStorageDetails findLogStorageDetails(
            @Nonnull ExecuteSpec executeSpec,
            @Nonnull String nodeClientId) {
        if (testConfig.getESpec().getLogStorage() == null)
            return null;
        for (ExperimentSetupConfig.ESpecLogStorageDetails logStorage: testConfig.getESpec().getLogStorage()) {
            if (logStorage.getPath() != null &&
                    executeSpec.getPath() != null &&
                    !executeSpec.getPath().equals(logStorage.getPath()))
                continue;
            if (logStorage.getNodeClientId() != null &&
                    !nodeClientId.equals(logStorage.getNodeClientId()))
                continue;
            return logStorage;
        }
        return null;
    }

    private boolean mustKeepLog(@Nonnull ExecuteSpec executeSpec,
                                @Nonnull String nodeClientId) {
        ExperimentSetupConfig.ESpecLogStorageDetails logStorage =
                findLogStorageDetails(executeSpec, nodeClientId);
        if (logStorage == null)
            return false;
        return logStorage.isStoreInFile() || logStorage.isStoreInLog();
    }

    private void keepLog(long logEventId,
                         @Nonnull ExecuteSpec executeSpec,
                         @Nonnull String nodeClientId,
                         @Nonnull String fullExecutablePath,
                         @Nullable String fullLogPath,
                         @Nonnull LimitedLiveLog currentLog) {
        String useablePath = executeSpec.getPath();
        if (useablePath == null) {
            useablePath = fullLogPath == null? null: fullLogPath.replaceAll(".*/", "");
        }
        if (useablePath == null) {
            useablePath = fullExecutablePath.replaceAll(".*/", "");
        }
        if (useablePath == null) {
            useablePath = "unknown-"+logEventId;
        }

        ExperimentSetupConfig.ESpecLogStorageDetails logStorage =
                findLogStorageDetails(executeSpec, nodeClientId);
        boolean addSeperate = logStorage != null && logStorage.isStoreInFile();
        boolean addInline = logStorage != null && logStorage.isStoreInLog();

        if (!addSeperate && !addInline) {
            //probably an error causing the log to be stored.
            addSeperate = true;
        }

        if (addSeperate) {
            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("log-" + useablePath + "-" + nodeClientId)
                            .setValue(currentLog.getText())
                            .setLarge(true/*currentLog.getReceivedSize() > 100*/)
                            .setHttpMediaTypeToXml()
            );
            currentTestResult.addLogLine(LogOutput.LogLineType.NOTE,
                    "Log content for "+useablePath+" @ "+nodeClientId+" has been stored. " +
                            "See \""+"log-" + useablePath + "-" + nodeClientId+"\"");
        }
        if (addInline) {
            currentTestResult.addLogLine(LogOutput.LogLineType.NOTE,
                    "Log content for "+useablePath+" @ "+nodeClientId+"\n\n"+
                            currentLog.getText(), true);
        }
    }

    public AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory getAutomaticUserAndSliceApiWrapperFactory() {
        return configInjector.getInstance(AutomaticUserAndSliceApiWrapper.BasicAutomaticUserAndSliceApiWrapperFactory.class);
    }

    private UserAndSliceApiWrapper userAndSliceApiWrapper;
    @Test(hardDepends = {"getEspec"}, groups = {"init"})
    public void findProject() {
        assertNotNull(user);
        userAndSliceApiWrapper =
                this.getAutomaticUserAndSliceApiWrapperFactory().create();

        switch (getTestConfig().getSlice().getProjectSource()) {
            case NONE: {
                project = null;
                break;
            }
            case PROVIDED: {
                project = getTestConfig().getSlice().getProject();
                break;
            }
            case AUTO: {
                project = CommonAMTest.findProject(
                        this,
                        this.getLogger(),
                        project,
                        user,
                        user.getUserUrn(),
                        userAndSliceApiWrapper,
                        true);
                break;
            }
        }

        if (project != null) {
            note("Will use project: \"" + project + "\"");
        } else {
            assertFalse(getTestConfig().getSlice().getFailIfNoProject(), "Empty project is not allowed by config");
            note("Will NOT use a project");
        }
    }

    NodeLoginTestStep nodeLoginTestStep;

    @Test(hardDepends = {"getEspec", "findProject"}, groups = {"init"})
    public void createExperiment() throws NoSuchAlgorithmException {
        assertNotNull(user);
        sliceName = CommonAMTest.createSliceName(
                this,
                userAuthorityServer,
                ApiInfo.getDefaultComponentManagerForEach(serversInRspec),
                "g",
                maxNodeNameLength,
                null);

        note("Will create slice \"" + sliceName + "\"");

        Instant requestedEndTime = Instant.now().plus(getTestConfig().getSlice().getExpireTimeMin(), ChronoUnit.MINUTES);
        note("Created slice (and slivers) will expire at " + requestedEndTime + "");

        //noinspection ConstantConditions
        this.nodeLoginTestStep =
                new NodeLoginTestStep(this, getTestConfig().generateNodeLoginTestStepConfig(geniUserProvider));

//        List<String> sshKeys = Collections.singletonList(KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey()));
        List<String> sshKeys = Collections.singletonList(nodeLoginTestStep.getSshKeyHelper().getSshPublicKeyString());
        UserSpec userSpec = new UserSpec(user.getUserUrnString(), sshKeys);
        List<UserSpec> userSpecs = Collections.singletonList(userSpec);

        @SuppressWarnings("ConstantConditions")
        String userPem = user == null || user.getPrivateKey() == null ? null : new String(KeyUtil.privateKeyToAnyPem(user.getPrivateKey()));

        experiment = new Experiment(
                sliceName,
                project,
                Collections.emptyList(),
                eSpecBundle,
                null,
                requestedEndTime,
                userSpecs,
                false,
                getRSpecGeneratorFactory(),
                new SingleSshGitAuthPreferences(userPem),
                false);

        experiment.addESpecLogListener(eSpecLogToTest);
    }

    @Test(hardDepends = {"createExperiment"}, groups = {"run"})
    public void runExperiment() {
        assertTrue(getTestConfig().getProvision().isEnabled(),
                "A disabled provision step is currently not supported.");

        notEnoughResourcesDetected = false;

        final ExperimentControllerFactory experimentControllerFactory =
                configInjector.getInstance(ExperimentControllerFactory.class);
        assertNotNull(experimentControllerFactory);

        experimentController =
                experimentControllerFactory.createExperimentController(experiment);
        assertNotNull(experimentController);

//        experimentController.setExperimentConnectivityTesterFactory(experimenterJobFactory::createTestConnectivityJob);
//        experimentController.addListener(bugReportExperimentControllerListener);

        experimentController.start();
        note("Started experimentController. ExperimentState is now " + experiment.getExperimentState());


        //TODO: if not enough free resources detected, do not wait
//        notEnoughResourcesDetected = todo;

        //Now wait.
        //This cannot be in a seperate method: we might miss calls in between the 2 methods otherwise (because experimentController is async).

        long maxTimeToWaitForReadyInMinutes = 0;
        long extraWaitTimeInSeconds = 0;
        if (!getTestConfig().getWaitForReady().isEnabled()) {
            //don't take this wait and extra time into account

            //consider the wait over
            waitForExperimentIsOver = true;
        }
        if (getTestConfig().getWaitForReady().getMaxTimeMin() > maxTimeToWaitForReadyInMinutes) {
            maxTimeToWaitForReadyInMinutes = getTestConfig().getWaitForReady().getMaxTimeMin();
        }
        if (getTestConfig().getWaitForReady().getExtraWaitTimeSeconds() > extraWaitTimeInSeconds) {
            extraWaitTimeInSeconds = getTestConfig().getWaitForReady().getExtraWaitTimeSeconds();
        }

        if (maxTimeToWaitForReadyInMinutes > 0) {
            note("Starting to wait for experiment(s) to become ready at " + new Date()+" (max wait is "+maxTimeToWaitForReadyInMinutes+" minutes)");
        } else {
            note("Will not wait for experiment(s) to become ready.");
        }

        long now = System.currentTimeMillis();
        long start = now;
        long deadline = now + (maxTimeToWaitForReadyInMinutes * 60 * 1000L);
        boolean allDone = false;
        boolean sawFailure = false;
        while (now < deadline && !allDone && !sawFailure) {
            if (!waitForExperimentIsOver) {
                assertTrue(getTestConfig().getWaitForReady().isEnabled(), "bug in runExperiment step code.");
                note("Current ExperimentState=" + experiment.getExperimentState());
                note("Checking if experiment is ready.");

                if (experiment.getSliceUrn() != null) {
                    sliceUrn = experiment.getSliceUrn();
                }

                if (experiment.getExperimentState() == ExperimentState.READY) {
                    note("Ready, because ExperimentState is READY");
                    waitForExperimentIsOver = true;
                }
                if (experiment.getExperimentState().isTerminated()) {
                    waitForExperimentIsOver = true;
                    sawFailure = true;
                    errorNonFatal("Failure, because ExperimentState is terminated");
                }
                if (!experiment.getExperimentState().isActiveState()) {
                    //TODO detect and set notEnoughResourcesDetected
                    waitForExperimentIsOver = true;
                    sawFailure = true;
                    errorNonFatal("Failure, because ExperimentState is not active state");
                }
//                if (experiment.getExperimentState() == TIMEOUT_WAITING) {
//                    waitForExperimentIsOver = true;
//                    sawFailure = true;
//                    errorNonFatal("Failure, because experiment timed out waiting to become ready");
//                }
            }


            allDone = waitForExperimentIsOver;

            if (!allDone && !sawFailure) {
                try {
                    Thread.sleep(30000 /*ms*/);
                } catch (InterruptedException e) { /* Ignore*/ }
            }
            now = System.currentTimeMillis();
        }
        long stop = now;


        if (maxTimeToWaitForReadyInMinutes > 0) {
            double waitTimeMinutes = (stop - start) / (60*1000L);
            note("Stopped waiting for experiment(s) to become ready at " + new Date()+" (Waited "+waitTimeMinutes+" minutes)");
        }

        if (extraWaitTimeInSeconds > 0) {
            note("Waiting " + extraWaitTimeInSeconds + " extra seconds before continuing. (due to resource.waitForReady.extraWaitTimeSeconds option)");
            try {
                Thread.sleep(extraWaitTimeInSeconds * 1000L);
            } catch (InterruptedException e) {
                warn("Extra waiting time was interrupted. This should not happen.", e);
            }
        }

        if (sawFailure) {
            errorFatal("One or more experiments have failed.");
        }

        if (!allDone) {
            errorFatal("One or more experiments did not become ready by the deadline.");
        }
    }

    public boolean isValidGeniRspec(String rspec) {
        return new BasicStringRspec(rspec).isValidRspec();
    }


    @Test(hardDepends = {"runExperiment"}, groups = {"run"})
    public void checkManifest() throws JFedException, IOException {
        if (notEnoughResourcesDetected) {
            warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
            return;
        }

        Slice slice = experiment.getSliceOrNull();
        assertNotNull(slice, "experiment slice should not be null");
        ManifestRspecSource manifestRspecSource = slice.getManifestRspec();
        assertNotNull(manifestRspecSource, "experiment slice manifestRspecSource should not be null");
        manifestRspecString =  manifestRspecSource.getRspecXmlString();
        assertNotNull(manifestRspecString, "experiment slice manifestRspecSource getRspecXmlString should not be null");

        @SuppressWarnings("ConstantConditions")
        String userPem = user == null || user.getPrivateKey() == null ? null : new String(KeyUtil.privateKeyToAnyPem(user.getPrivateKey()));
        String requestRspec =
                new String(new FileFetcher(
                        eSpec.getRspecs().get(0).getSource(),
                        eSpecBundle,
                        getRSpecGeneratorFactory(),
                        new SingleSshGitAuthPreferences(userPem)).fetchBytes(),
                        StandardCharsets.UTF_8);

        boolean validRequest = isValidGeniRspec(requestRspec);
        validManifest = isValidGeniRspec(manifestRspecString);

        //noinspection ConstantConditions
        if (requestRspec != null && validRequest && validManifest) {
            setErrorsNotFatal();
            assertTrue(CommonAMTest.sameNodesInRequestAndManifest(this, requestRspec, manifestRspecString));
            setErrorsFatal();
        }

        if (validManifest) {
            boolean foundNodeLogin = nodeLoginTestStep.parseSshInfoFromGeni3ManifestRspec(manifestRspecString, null, null);
            if (foundNodeLogin)
                note("Successfully found node login in manifest");
            else
                note("Did not find node login info in manifest");

            if (getTestConfig().getWaitForReady().getSkipFinalManifestCheck())
                nodeLoginTestStep.checkManifestCorrectness(manifestRspecString, testedAuthority);
        }

    }

    @Test(softDepends = {"runExperiment", "checkManifest"},
            hardDepends = {"getEspec", "createExperiment"},
            groups = {"cleanup", "run"})
    public void deleteExperiment() {
        if (getTestConfig().getCleanup().isEnabled()) {
            note("Will call experimentController.stop();");
            experimentController.stop();
        } else {
            note("Experiment Cleanup disabled");
            deleted = true;
        }


        long maxTimeToWaitForDeleteInMinutes = 0;
        if (getTestConfig().getCleanup().getMaxWaitTimeMinutes() > maxTimeToWaitForDeleteInMinutes) {
            maxTimeToWaitForDeleteInMinutes = getTestConfig().getCleanup().getMaxWaitTimeMinutes();
        }


        note("Will wait until experiment is terminated");
        long now = System.currentTimeMillis();
        long deadline = now + (maxTimeToWaitForDeleteInMinutes * 60 * 1000L);
        boolean allDeleted = false;
        while (now < deadline && !deleted) {
            if (!deleted) {
                note("Current ExperimentState=" + experiment.getExperimentState());
                if (experiment.getExperimentState() == ExperimentState.EMPTY ||
                        experiment.getExperimentState() == ExperimentState.EXPIRED ||
                        experiment.getExperimentState() == ExperimentState.FAILED) {
                    note("Experiment was terminated");
                    deleted = true;
                }
            }

            if (!deleted) {
                try {
                    Thread.sleep(30000 /*ms*/);
                } catch (InterruptedException e) { /* Ignore*/ }
            }
            now = System.currentTimeMillis();
        }

        if (!deleted) {
            errorFatal("Experiment(s) did not terminate within " + maxTimeToWaitForDeleteInMinutes + " minutes. Gave up waiting for it.");
        } else {
            note("All experiments deleted");
        }
    }

    private class ESpecLogToTest implements ESpecLogListener {

        @Override
        public void onPreFileLoad() {
            note("ExperimentSpecification: Start loading files.");
        }

        @Override
        public void onPostFileLoad(@Nonnull FileSource fileSource, boolean isFast, long byteSize, @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                errorNonFatal("ExperimentSpecification: Failed to load file \""+fileSource.getBasename()+"\"");
            else {
                if (!isFast)
                    note("ExperimentSpecification: Successfully loaded file \""+fileSource.getBasename()+"\" ("+byteSize+" byte)");
            }
        }

        @Override
        public void onPostFileLoadAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification: Failed to load all files.");
            else
                note("ExperimentSpecification: Successfully finished loading files.");
        }

        @Override
        public void onPreRSpec() {
            note("ExperimentSpecification: Start Provisioning RSpec.");
        }

        @Override
        public void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec) {
            note("ExperimentSpecification: onRequestRSpecKnown length="+requestRspec.length());
        }

        @Override
        public void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status) {

        }

        @Override
        public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) {
            if (!success)
                errorNonFatal("ExperimentSpecification: Failed to Provision RSpec");
            else
                note("ExperimentSpecification: Succesfully Provisioned RSpec with nodes "+allNodeClientIds.stream().collect(Collectors.joining()));
        }

        @Override
        public void onPreDir() {
            note("ExperimentSpecification: Start dir setup.");
        }

        @Override
        public void onPreDirNode(@Nonnull String nodeClientId) {

        }

        @Override
        public void onPreDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId) {

        }

        @Override
        public void onPostDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to setup dir at \""+fullPath+"\"");
            else
                note("ExperimentSpecification: (on \"" + nodeClientId + "\") Succesfully setup dir at \""+fullPath+"\"");
        }

        @Override
        public void onPostDirNode(@Nonnull String nodeClientId, boolean success) {

        }

        @Override
        public void onPostDirAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification: Failed to setup dirs");
            else
                note("ExperimentSpecification: Succesfully setup dirs");
        }

        @Override
        public void onPreUpload() {
            note("ExperimentSpecification: Start upload(s).");
        }

        @Override
        public void onPreUploadNode(@Nonnull String nodeClientId) {

        }

        @Override
        public void onPreUpload(long logEventId, @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull UploadProgressTracker uploadProgressTracker) {

        }

        @Override
        public void onPostUpload(long logEventId, @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath,
                                 @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to upload file \""+uploadSpec.getDesc()+"\" at \""+fullPath+"\"");
            else
                note("ExperimentSpecification: (on \"" + nodeClientId + "\") Succesfully uploaded file \"" + uploadSpec.getDesc() + "\" at \""+fullPath+"\"");
        }

        @Override
        public void onPostUploadNode(@Nonnull String nodeClientId, boolean success) {

        }

        @Override
        public void onPostUploadAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification: Failed to upload at least one file to one node");
            else
                note("ExperimentSpecification: Succesfully uploaded files");
        }

        @Override
        public void onPreExecute() {
            note("ExperimentSpecification: Start execute(s).");
        }

        private final Map<Long, LimitedLiveLog> tmpLogs = new HashMap<>();
        @Override
        public void onPreExecute(long logEventId,
                                 @Nonnull ExecuteSpec executeSpec,
                                 @Nonnull String fullExecutablePath,
                                 @Nullable String fullLogPath,
                                 @Nonnull String nodeClientId,
                                 @Nonnull LimitedLiveLog currentLog) {
            if (mustKeepLog(executeSpec, nodeClientId)) {
                keepLog(logEventId, executeSpec, nodeClientId, fullExecutablePath, fullLogPath, currentLog);
            } else {
                tmpLogs.put(logEventId, currentLog);
            }
        }

        @Override
        public void onPostExecute(long logEventId,
                                  @Nonnull ExecuteSpec executeSpec,
                                  @Nonnull String fullPath,
                                  @Nullable String remoteLogPath,
                                  @Nonnull String nodeClientId,
                                  int exitStatus, @Nonnull ESpecStepStatus status) {
            if (status.isFailure()) {
                errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to execute script \"" + executeSpec.getDesc() + "\" at \"" + fullPath + "\" exitStatus=" + exitStatus);
                keepLog(logEventId, executeSpec, nodeClientId, fullPath, remoteLogPath, tmpLogs.get(logEventId));
            } else
                note("ExperimentSpecification: (on \"" + nodeClientId + "\") Succesfully executed script \"" + executeSpec.getDesc() + "\" at \""+fullPath+"\" exitStatus="+exitStatus);

            tmpLogs.remove(logEventId);
        }

        @Override
        public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification: Failed to execute script \"" + executeSpec.getDesc() + "\" on at least one node");
            else
                note("ExperimentSpecification: Succesfully executed script \"" + executeSpec.getDesc() + "\" on all nodes");
        }

        @Override
        public void onPostExecuteAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification: Failed to execute at least one script on at least one node");
            else
                note("ExperimentSpecification: Succesfully executed all scripts");
            tmpLogs.clear();
        }

        @Override
        public void onPreAnsible() {
            note("ExperimentSpecification: Start ansible(s).");
        }

        @Override
        public void onPreAnsiblePlaybook(long logEventId, @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec, @Nonnull String fullPlaybookPath, @Nullable String fullLogPath, @Nonnull String nodeClientId, @Nonnull LimitedLiveLog currentLog) {

        }

        @Override
        public void onPostAnsiblePlaybook(long logEventId,
                                          @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                          @Nonnull String fullPath,
                                          @Nonnull String remoteLogPath,
                                          @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to execute ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" at \""+fullPath+"\"");
            else
                note("ExperimentSpecification: (on \"" + nodeClientId + "\") Succesfully executed ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" at \""+fullPath+"\"");
        }

        @Override
        public void onPostAnsibleAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification: Failed to ansible at least one script on at least one node");
            else
                note("ExperimentSpecification: Succesfully ansibled all scripts");
        }
    }
}
